/*
Copyright (C) Cedilla Software Services Ltd. 1999. License is granted to use this software subject to
the following conditions:-

Preamble:

The intent of this document is to state the conditions under which a Package may be copied, 
such that the Copyright Holder maintains some semblance of artistic control over the development 
of the package, while giving the users of the package the right to use and distribute the 
Package in a more-or-less customary fashion, plus the right to make reasonable modifications.

Definitions:

"Package" refers to the collection of files distributed by the Copyright Holder, and 
derivatives of that collection of files created through textual modification.

"Standard Version" refers to such a Package if it has not been modified, or has been modified 
in accordance with the wishes of the Copyright Holder.

"Copyright Holder" is whoever is named in the copyright or copyrights for the package.

"You" is you, if you're thinking about copying or distributing this Package.

"Reasonable copying fee" is whatever you can justify on the basis of media cost, duplication 
charges, time of people involved, and so on. (You will not be required to  justify it to 
the Copyright Holder, but only to the computing community at large as a market that must bear the fee.)

"Freely Available" means that no fee is charged for the item itself, though there may be 
fees involved in handling the item. It also means that recipients of the item may redistribute 
it under the same conditions they received it.

1. You may make and give away verbatim copies of the source form of the Standard Version of this 
Package without restriction, provided that you duplicate all of the original copyright 
notices and associated disclaimers.

2. You may apply bug fixes, portability fixes and other modifications derived from the Public 
Domain or from the Copyright Holder. A Package modified in such a way shall still be considered 
the Standard Version.

3. You may otherwise modify your copy of this Package in any way, provided that you 
insert a prominent notice in each changed file stating how and when you changed that file, and provided 
that you do at least ONE of the following:

a) place your modifications in the Public Domain or otherwise make them Freely Available, 
such as by posting said modifications to Usenet or an equivalent medium, or placing the 
modifications on a major archive site such as ftp.uu.net, or by allowing the 
Copyright Holder to include your modifications in the Standard Version of the Package.

b) use the modified Package only within your corporation or organization.

c) rename any non-standard executables so the names do not conflict with 
standard executables, which must also be provided, and provide a separate manual page for 
each non-standard executable that clearly documents how it differs from the Standard Version.

d) make other distribution arrangements with the Copyright Holder.

4. You may distribute the programs of this Package in object code or executable 
form, provided that you do at least ONE of the following:

a) distribute a Standard Version of the executables and library files, together with 
instructions (in the manual page or equivalent) on where to get the Standard Version.

b) accompany the distribution with the machine-readable source of the Package 
with your modifications.

c) accompany any non-standard executables with their corresponding Standard Version 
executables, giving the non-standard executables non-standard names, and clearly documenting 
the differences in manual pages (or equivalent), together with instructions on where to 
get the Standard Version.

d) make other distribution arrangements with the Copyright Holder.

5. You may charge a reasonable copying fee for any distribution of this Package. You may 
charge any fee you choose for support of this Package. You may not charge a fee for
this Package itself. However, you may distribute this Package in aggregate with other 
(possibly commercial) programs as part of a larger (possibly commercial) software
distribution provided that you do not advertise this Package as a product of your own.

6. The scripts and library files supplied as input to or produced as output from the 
programs of this Package do not automatically fall under the copyright of this Package, 
but belong to whomever generated them, and may be sold commercially, and may be 
aggregated with this Package.

7. C, Java or perl subroutines supplied by you and linked into this Package shall not be 
considered part of this Package.

8. The name of the Copyright Holder may not be used to endorse or promote products 
derived from this software without specific prior written permission.

9. THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS 
FOR A PARTICULAR PURPOSE.

The End

Acknowledgments:

My thanks to the following for their suggestions and help:

Zubair Quraishi
*/
import java.io.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.Enumeration;
import com.sun.xml.parser.Resolver;
import com.sun.xml.tree.XmlDocument;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class Comb
{
    private String _copyrightString = null;
    private Hashtable _classes = new Hashtable();
    private Hashtable _codeFragments = new Hashtable();

    public Comb(XmlDocument doc) {
	
	    Element root = doc.getDocumentElement();
        Element child = (Element)root.getFirstChild();
        
        if (child.getTagName().equals("COPYRIGHT")) {
            preProcessCopyright(child);
            child = (Element)child.getNextSibling();
        }
        
        //Gather all the classes and code frags
        while (child != null) {
            preProcessSection(child);
            child = (Element)child.getNextSibling();
        }
        
        System.out.println("Comb: processing");
        
        //Build a code fragment tree
        for (Enumeration e = _codeFragments.keys(); e.hasMoreElements();) {
            CodeFragment fragment = (CodeFragment)_codeFragments.get(e.nextElement());
            String extendsStr = fragment.getExtends();
            
            if (extendsStr != null) {
                if (_codeFragments.containsKey(extendsStr)) {
                    CodeFragment extendedFragment = (CodeFragment)_codeFragments.get(extendsStr);
                    extendedFragment.addFragment(fragment.getName());
                } else if (_classes.containsKey(extendsStr)) {
                    ClassRep classRep = (ClassRep)_classes.get(extendsStr);
                    classRep.addFragment(fragment.getName());
                } else {
                    System.out.println("Comb: unable to find extended fragment " + extendsStr + " mentioned in fragment " + fragment.getName());
                }
            }
        }
        
        System.out.println("Comb: outputing");
        for (Enumeration e = _classes.elements(); e.hasMoreElements();) {
            ClassRep classRep = (ClassRep)e.nextElement();
            classRep.output();
        }
    }
    
    //
    // Reading and writing an XML document stored in a file.
    //
    public static void main (String argv [])
    {
	InputSource	input;
	XmlDocument	doc;
	
	if (argv.length != 1) {
	    System.err.println ("Usage: cmd filename");
	    System.exit (1);
	}

	try {
	    // turn the filename into an input source
	    input = Resolver.createInputSource (new File (argv [0]));

	    // turn it into an in-memory object
	    // ... the "false" flag says not to validate
	    doc = XmlDocument.createXmlDocument (input, true);

	    // normalize text representation
	    doc.getDocumentElement ().normalize ();
	    
	    // Comb through the file to create a java file
	    Comb comb = new Comb(doc);

	} catch (SAXParseException err) {
	    System.out.println ("** Parsing error" 
		+ ", line " + err.getLineNumber ()
		+ ", uri " + err.getSystemId ());
	    System.out.println("   " + err.getMessage ());
	    // print stack trace as below

	} catch (SAXException e) {
	    Exception	x = e.getException ();

	    ((x == null) ? e : x).printStackTrace ();

	} catch (Throwable t) {
	    t.printStackTrace ();
	}

	System.exit (0);
    }
    
    private void preProcessCopyright(Element code) {
        Text text = (Text)code.getFirstChild();
        try {
                _copyrightString = text.getData();
                System.out.println("Comb: Copyright");
        } catch (Exception e) {
            System.out.println(e);
        }
    }
        

    private void preProcessSection(Element section) {
        Element child = (Element)section.getFirstChild();
        
        //Skip Narrative
        
        while (child != null) {
            if (child.getTagName().equals("NARRATIVE")) {
                //skip
            } else if (child.getTagName().equals("CODE")) {
                preProcessCode(child);
            } else if (child.getTagName().equals("CLASS")) {
                preProcessClass(child);
            } else if (child.getTagName().equals("INTERFACE")) {
                preProcessInterface(child);
            } else if (child.getTagName().equals("FILE")) {
                preProcessFile(child);
            } else if (child.getTagName().equals("SUBSECTION")) {
                preProcessSubSection(child);
            }
            child = (Element)child.getNextSibling();
        }
    }

    private void preProcessSubSection(Element subsection) {
        Element child = (Element)subsection.getFirstChild();
        
        //Skip Narrative
        
        while (child != null) {
            if (child.getTagName().equals("NARRATIVE")) {
                //skip
            } else if (child.getTagName().equals("CODE")) {
                preProcessCode(child);
            } else if (child.getTagName().equals("CLASS")) {
                preProcessClass(child);
            } else if (child.getTagName().equals("INTERFACE")) {
                preProcessInterface(child);
            } else if (child.getTagName().equals("FILE")) {
                preProcessFile(child);
            }
            child = (Element)child.getNextSibling();
        }
    }
    
    private void preProcessCode(Element code) {
        String name = code.getAttribute("NAME");
        String extendsStr = code.getAttribute("EXTENDS");
        String weightStr = code.getAttribute("WEIGHT");
        int weight = 0;
        
        if (weightStr.length() > 0) {
            try {
                weight = Integer.parseInt(weightStr);
            } catch (NumberFormatException e) {
                System.out.println("Comb: can't parse weight " + weightStr + " in fragment " + name);
            }
        }
        
        Text text = (Text)code.getFirstChild();
        try {
            String codeString = text.getData();
            
            if (extendsStr != null) {
                CodeFragment fragment = new CodeFragment();
                fragment.setName(name);
                fragment.setCode(codeString);
                fragment.setExtends(extendsStr);
                fragment.setWeight(weight);
                _codeFragments.put(name, fragment);
                System.out.println("Comb: Adding fragment " + name);
            }
        } catch (Exception e) {
            System.out.println(e);
        }
    }
    
    private void preProcessClass(Element cls) {
        ClassRep classRep = new ClassRep(ClassRep.CLASS);
        String name = cls.getAttribute("NAME");
        classRep.setName(name);
        classRep.setExtends(cls.getAttribute("EXTENDS"));
        classRep.setImplements(cls.getAttribute("IMPLEMENTS"));
        classRep.setPackage(cls.getAttribute("PACKAGE"));
        classRep.setAbstract(cls.getAttribute("ABSTRACT").equals("T"));
        
        _classes.put(name, classRep);
        
        CodeFragment prelude = new CodeFragment();
        prelude.setName(name + ".prelude");
        prelude.setCode("");
        _codeFragments.put(name + ".prelude", prelude);
        
        System.out.println("Comb: Adding class " + name);
        
    }
    
    private void preProcessInterface(Element inter) {
        ClassRep classRep = new ClassRep(ClassRep.INTERFACE);
        String name = inter.getAttribute("NAME");
        classRep.setName(name);
        classRep.setExtends(inter.getAttribute("EXTENDS"));
        classRep.setPackage(inter.getAttribute("PACKAGE"));
        classRep.setImplements("");
        
        _classes.put(name, classRep);
        
        CodeFragment prelude = new CodeFragment();
        prelude.setName(name + ".prelude");
        prelude.setCode("");
        _codeFragments.put(name + ".prelude", prelude);
        
        System.out.println("Comb: Adding interface " + name);
    }
    
    private void preProcessFile(Element file) {
        ClassRep classRep = new ClassRep(ClassRep.FILE);
        String name = file.getAttribute("NAME");
        classRep.setName(name);
        
        Text text = (Text)file.getFirstChild();
        try {
            String fileString = text.getData();
            classRep.setFileContents(fileString);
        } catch (Exception e) {
            System.out.println(e);
        }
        
        _classes.put(name, classRep);
        
        System.out.println("Comb: Adding file " + name);        
    }
    
    
    
    class ClassRep {
        //Static Members//////////////////////////////////////////////////////////////////
        public static final int CLASS = 0;
        public static final int INTERFACE = 1;
        public static final int FILE = 2;
        
        //Member Variables////////////////////////////////////////////////////////////////
        private int _type = CLASS;
        private String _name = null;
        private String _extends = null;
        private String _implements = null;
        private String _package = null;
        private boolean _abstract = false;
        private String _fileContents = null;
        private ArrayList _fragments = new ArrayList();
        
        //Constructors////////////////////////////////////////////////////////////////////
        public ClassRep(int type) {
            _type = type;
        }
        
        public void setType(int type) {
            _type = type;
        }
        
        public int getType() {
            return _type;
        }
        
        public void setName(String name) {
            _name = name;
        }
        
        public String getName() {
            return _name;
        }

        public void setExtends(String ext) {
            _extends = ext;
        }
        
        public String getExtends() {
            return _extends;
        }

        public void setImplements(String imp) {
            _implements = imp;
        }
        
        public String getImplements() {
            return _implements;
        }
        
        public void setPackage(String pack) {
            _package = pack;
        }
        
        public String getPackage() {
            return _package;
        }
        
        public void setAbstract(boolean abs) {
            _abstract = abs;
        }
        
        public boolean isAbstract() {
            return _abstract;
        }
                
        public void setFileContents(String str) {
            _fileContents = str;
        }
        
        public String getFileContents() {
            return _fileContents;
        }
        
        public void addFragment(String fragmentName) {
            _fragments.add(fragmentName);
        }
        
        public void output() {
            FileOutputStream fos; 
            PrintStream pos = null;

	        try {
	            fos = new FileOutputStream(getName() + (_type != FILE ? ".java" : ""));
	            pos = new PrintStream(fos);
        	    
	        } catch (IOException e) {
	            System.out.println(e);
	        }

	        if (_type != FILE) {
	            outputCopyright(pos);
	            
	            pos.println("//This file generated automatically from the MARIUS LP system, please do not edit.");
    	        
	            if (_package.length() > 0) {
	                pos.println("package " + _package + ";");
	                pos.println();
	            }
    	        
	            CodeFragment prelude = (CodeFragment)_codeFragments.get(_name + ".prelude");
	            prelude.output(pos);
	            pos.println();
	        
	            pos.print("public ");
    	        
	            if (isAbstract()) {
	                pos.print("abstract ");
	            }
    	        
	            pos.print((_type == INTERFACE ? "interface " : "class ") + getName());
	            if (_extends.length() > 0) {
	                pos.print(" extends " + _extends);
	            }
    	        
	            if (_implements.length() > 0) {
	                pos.print(" implements " + _implements);
	            }
    	        
	            pos.println(" {");
    	        
	            outputCode(pos);
    	        
	            pos.println();
	            pos.println("} //End of class " + _name);
	        } else { //File output
	            if (_fileContents != null) {
	                pos.print(_fileContents);
	            }
	            outputCode(pos);
	        }
	    
        }
        
        private void outputCode(PrintStream pos) {
            Collections.sort(_fragments, new WeightComparator());
            for (int i = 0; i < _fragments.size(); i++) {
                String fragmentName = (String)_fragments.get(i);
                CodeFragment fragment = (CodeFragment)_codeFragments.get(fragmentName);
                fragment.output(pos);
            }
        }
        
        private void outputCopyright(PrintStream pos) {
            if (_copyrightString != null) {
                pos.println(_copyrightString);
            }
        }
        
    } //End of class classrep
    
    public class CodeFragment {
        private String _name = null;
        private String _extends = null;
        private String _code = null;
        private int _weight;
        
        private ArrayList _fragments = new ArrayList();

        public void setName(String name) {
            _name = name;
        }
        
        public String getName() {
            return _name;
        }

        public void setExtends(String ext) {
            _extends = ext;
        }
        
        public String getExtends() {
            return _extends;
        }

        public void setCode(String code) {
            _code = code;
        }
        
        public String getCode() {
            return _code;
        }

        public void setWeight(int weight) {
            _weight = weight;
        }
        
        public int getWeight() {
            return _weight;
        }
        
        public void addFragment(String fragmentName) {
            _fragments.add(fragmentName);
        }
        
        public void output(PrintStream pos) {
            pos.print(getCode());
            Collections.sort(_fragments, new WeightComparator());            
            for (int i = 0; i < _fragments.size(); i++) {
                String fragmentName = (String)_fragments.get(i);
                CodeFragment fragment = (CodeFragment)_codeFragments.get(fragmentName);
                fragment.output(pos);
            }
        }
        
    }//End of class codeFragment
    
    public class WeightComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            String fragName1 = (String) o1;
            String fragName2 = (String) o2;
            CodeFragment f1 = (CodeFragment)_codeFragments.get(fragName1);
            CodeFragment f2 = (CodeFragment)_codeFragments.get(fragName2);
            
            return f1.getWeight() - f2.getWeight();
        }
    }//End class WeightComparator
        
}
